/******************************************************************************
        Author:         Stephen Strudwick
        Department:     Unix Programming Group
        Modify Date:    09/01/2004
        Company:        Pipex Communications

        Description:    Blowfish util functions.

                        Provides functionality for an encrypted connection over a 
						TCP/IP socket using blowfish.

						Provides shared keys on a per user basis.

						username/password authentication has been removed to keep
						it simple.

						int bf_client_handshake(int socket, BFkey_type *ks);
						int bf_server_handshake(int socket, BFkey_type *ks);

						int bf_sendall(int socket, BFkey_type *ks, char *buf, int bytes);
						int bf_recvcall(int socket, BFkey_type *ks, char *vptr, int bytes);

******************************************************************************/

#include "bf_utils.h"

/*****************************************************************************
 Functions
******************************************************************************/

/*****************************************************************************
 Function name : bf_client_handshake
 Description   : Client Performs a hanshake using the default key
 Input Values  : 
 Return Values : 0 on success
                 Negative on error
******************************************************************************/

int bf_client_handshake(int socket, BFkey_type *ks) {
  char buf[HANDSHAKE_SIZE+1];

  /* Handshake */
  if (bf_recvall(socket, ks, buf, HANDSHAKE_SIZE) <= 0)
    return ERROR_READLINE;

  /* NULL terminate it */
  buf[HANDSHAKE_SIZE] = '\0';

  strrev(buf);
  bf_sendall(socket, ks, buf, HANDSHAKE_SIZE);

  return 0;
}

/*****************************************************************************
 Function name : bf_server_handshake
 Description   : server side handshake
 Inputs        : int socket, BFkey_type *ks
 Output        : int (0 success or negative failure)
******************************************************************************/

int bf_server_handshake(int socket, BFkey_type *ks) {
  if (!handshake(socket, ks)) {
    syslog(LOG_CRIT, "CRIT: bad blowfish handshake");
    return ERROR_HANDSHAKE;
  } else {
    return 0;
  }
}

/*****************************************************************************
 Function name : Read
 Description   : Re-implementation of read to work with encryption.
                 In encrypted mode reads MAX_STREAM_CHUNK bytes at a time,
                 blocking otherwise and decrypts the received blocks.
 Inputs        : 
 Output        : int
******************************************************************************/

int Read(int socket, BFkey_type *ks, char *buf) {
  int rc;
  fd_set fds;
  struct timeval tv;
  int read_cnt = 0;

  char *decrypted;

  tv.tv_sec = TIMEOUT;
  tv.tv_usec = 0;
  FD_ZERO( &fds );
  FD_SET( socket, &fds );

  if (read_cnt <= 0) {
    again:
    rc = select( socket+1, &fds, 0, 0, &tv );
      if( rc < 0 ) {
        return -1;
      } else if (rc ==0) {
        return ERROR_READLINE;
      } else {
      if ((read_cnt = recv_waitall(socket, buf, MAX_STREAM_CHUNK)) < 0) {
        if(errno == EINTR)
          goto again;
         return -1;
      } else if (read_cnt == 0)
        return 0;
    }
  }

  buf[read_cnt] = NUL;

  decrypted = blowfish_decrypt(ks, buf, read_cnt);
  if (decrypted) {
    memcpy(buf,decrypted,MAX_STREAM_CHUNK);
    free(decrypted);
  } else {
    return ERROR_ENCRYPTION;
  }

  return read_cnt;
}


int recv_waitall(SOCKET fd, char *buf, int size) {
  int rc = 0;

    rc = recv(fd, buf, size, 0);

	if (rc < 0) {
      return -1;
    } else if (rc == 0) {
      return 0;
    } else if (rc < size) {
	  return rc + recv_waitall(fd, &buf[rc],size-rc);
	}

  return rc;

}

/*****************************************************************************
 Function name : bf_recvall
 Description   : Reads a line of encrypted data from a socket. This functions
	       		 reads in data until len bytes are read, it breaks on nothing
				 else.

                 buf is populated with data read.

				 The data can be of any format.

 Inputs        : int socket, BFkey_type *ks, char *buf, int len
 Output        : int (Number of chars read)
******************************************************************************/

int bf_recvall(int socket, BFkey_type *ks, char *buf, int len) {
  char *ptr;
  char vptr[MAX_STREAM_CHUNK+1];
  int tc = 0;
  int rc = 0;
  buf[0] = '\0';
  ptr = buf;

  /* clear the receive buffer */
  bzero(buf,len);

  while(tc < len) {
    if ((rc = Read(socket, ks, vptr)) >= 1) {

      tc += rc;
      if (tc <= len) {
	memcpy(ptr,vptr,MAX_STREAM_CHUNK);
	ptr += MAX_STREAM_CHUNK;
      } else {
	tc = tc - (tc-len);
	memcpy(ptr,vptr,MAX_STREAM_CHUNK - (tc-len));
        break;
      }

    } else if (rc == 0) {
      if (tc == 1)
        return 0;
      else
        break;
    } else
      return -1;
  }

  return (tc);
}

/*****************************************************************************
 Function name : bf_sendall
 Description   : Send a line of encrypted data to the client.
 Inputs        : int socket, BFkey_type *ks, char *buf, int bytes
 Output        : int (0 on success, negative on failure)
******************************************************************************/

int bf_sendall(int socket, BFkey_type *ks, char *buf, int bytes) {
  int total=0;
  int bytesleft=0;
  int n=0;
  c_string *result;

  /* Encrypt it */
  result = (c_string *)Malloc(sizeof(*result));
  blowfish_crypt(ks, buf, result, bytes);
  bytesleft=result->len;

  /* send all the data */
  while(total<result->len){

    /* send some data */
    n=send(socket,result->c_string+total,bytesleft,0);

    /* break on error */
    if(n==-1)
      break;

    /* apply bytes we sent */
    total+=n;
    bytesleft-=n;
  }

  free(result->c_string);
  free(result);

  /* return -1 on failure, 0 on success */
  return n==-1?-1:0;
}

/*****************************************************************************
 Function name : handshake
 Description   : creates a random string, sends it to client and waits for
                 for it to be sent back reversed. All this is done encrypted
				 with a shared key.
 Return Values : true or false
******************************************************************************/

int handshake(int socket, BFkey_type *ks) {
  char handshake[HANDSHAKE_SIZE+1];
  char response[HANDSHAKE_SIZE+1];
  int len;

  /* Generate random character line of text */
  make_random_line(handshake, HANDSHAKE_SIZE);

  /* Send to client (client reverses it and sends back) */
  bf_sendall(socket, ks, handshake, HANDSHAKE_SIZE);

  /* Get it back */
  len = bf_recvall(socket, ks, response, HANDSHAKE_SIZE);
  if(len <= 0) {
    syslog(LOG_WARNING, "WARNING: Error getting line from connection");
    return(0);
  }

  /* NULL terminate it */
  response[HANDSHAKE_SIZE] = '\0';

  /* Reverse it */
  /* Verify */
  if(strncmp(handshake, _strrev(_strdup(response)), HANDSHAKE_SIZE) == 0)
    return 1;
  else
    return 0;
}

/*****************************************************************************
 Function name : make_random_line
 Description   : makes a random string of alphanumeric chars
 Return Values : void, result returned in input string
******************************************************************************/

void make_random_line(char *line, int len) {
  int n,j;
  char salt_list[1024];
  strncpy(salt_list, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789", 62);

  srand(_getpid());
  for(n=0;n<len;n++) {
    j=1+(int) (61.0*rand()/(RAND_MAX+1.0));
    line[n] = salt_list[j];
  }
  line[n] = '\r'; 
  line[n+1] = '\n'; 
  line[n+2] = '\0'; 
}
